/* mem_over.c - Provides the overflow capabilities of NJAMD
 *
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 */
#include <lib/njamd.h>
#include <lib/libc.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <errno.h>

/* FUNCTION: __nj_overflow_alloc
 * SUMMARY: General purpose allocator, cuts memory off of the slab, protects
 *          the page after it. In order for the users memory to end on this
 *          page, we must offsett it as well.
 */
void *__nj_none_alloc(void *ptr, size_t len, int fill, u_long *laddr, char *caller)
{
	struct heap_entry *inf;
	size_t offset;
	void *buf;

	if(!len && NJ_ISSET(__nj_flags, NJ_ALLOW_MALLOC0))
	{
		return NULL;
	}

	/* Ask a stupid question, get a stupid answer */
	if(len <= 0)
	{
		__nj_eprintf("NJAMD/%s: Request must be > 0 (%d)\n", caller, len);
		__nj_print_stack_trace(laddr, TRACE_DEPTH);
		errno = EINVAL;
		return NULL;
	}

	offset = NJ_MAX(ALIGNMENT, sizeof(void *));

	len = ALIGN_TO(len, ALIGNMENT);

	if(!__nj_libc_realloc || !__nj_libc_malloc || !__nj_libc_calloc)
	{
		__nj_eprintf("NJAMD_PROT=none can't be used on this system (no "
			" dlopen?)\n");
		exit(1);
	}

	TR;
	/* Realloc? */
	if(ptr != NULL)
	{
		inf = *(struct heap_entry **)(ptr-offset);
		TR;

		/* Report the old memory as freed */	
		__nj_update_stats(-inf->ulen, -(NJ_MAX(ALIGNMENT, sizeof(void *)) + inf->ulen + sizeof(u_long)), -(NJ_MAX(ALIGNMENT, sizeof(void *)) + inf->ulen + sizeof(u_long))); 
		
		buf = __nj_libc_realloc(ptr-offset, len + offset + sizeof(u_long)) + offset;
	}
	else
	{
		TR;
		if(fill != -1)
			buf = __nj_libc_calloc(1, len + offset + sizeof(u_long));
		else
			buf = __nj_libc_malloc(len + offset + sizeof(u_long));
		TR;
		*(struct heap_entry **)buf = inf = __nj_new_heap_entry(buf, len, laddr);
		
		
		buf += offset;
	}
	TR;

	__nj_update_stats(len, offset + len + sizeof(u_long), offset + len + sizeof(u_long)); 

	inf->ulen = len;
	memcpy(inf->loc.alloced, laddr, TRACE_DEPTH*sizeof(u_long));
	inf->u_addr.start = buf;
	
	TR;
	*(u_long *)(buf + len) = NJ_FENCEPOST;

	return buf;
}

/* FUNCTION: __nj_none_free
 * SUMMARY: Uses standard libc free with some boundry checks. Does not record
 * what was freed when. 
 */ 
void __nj_none_free(void *buf, u_long *ret)
{
	struct heap_entry *inf;
	void *start;

	if(buf == NULL || __nj_known_chk_free == NO_FREE)
	{
		return;
	}

	TR;
	
	start = buf - NJ_MAX(ALIGNMENT, sizeof(void *));

	inf = *(struct heap_entry **)start;

	TR;
	
	if(!(__nj_heap_tbl <= inf && inf <= __nj_heap_top))
	{
		__nj_eprintf("NJAMD free: Heap corruption. Try using underflow option\n");
		raise(SIGSEGV);
		exit(1);
	}

	TR;
	
	if(*(u_long *)(buf + inf->ulen) != NJ_FENCEPOST)
	{
		__nj_eprintf("NJAMD free: Heap corruption. Try using overflow option\n");
		raise(SIGSEGV);
		exit(1);
	}
	
	__nj_update_stats(-inf->ulen, -(NJ_MAX(ALIGNMENT, sizeof(void *)) + inf->ulen + sizeof(u_long)), -(NJ_MAX(ALIGNMENT, sizeof(void *)) + inf->ulen + sizeof(u_long))); 
	
	TR;	
	__nj_free_stack_push(inf);
	
	TR;
	if(__nj_libc_free)
		__nj_libc_free(start);
	else
	{
		__nj_eprintf("NJAMD_PROT=none can't be used on this system (no "
			" dlopen?)\n");
		exit(1);
	}
}

/* FUNCTION: __nj_none_stats
 * SUMMARY: This is just a quick little stat function so I can see how much
 * memory is mapped. 
 */
void __nj_none_stats(void)
{
	struct heap_entry *p;
	int user = 0;
	int tot_used = 0;

	for(p = __nj_heap_tbl; p < __nj_heap_top; p++)
	{
		if(p->ulen != 0)
		{
			if(!entry_is_freed(p))
			{
				user += p->ulen;
				tot_used += p->ulen + NJ_MAX(ALIGNMENT, sizeof(void *)) + sizeof(u_long);
			}
		}
	}

	__nj_eprintf("\nNJAMD totals:\n\tUser memory:\t%d bytes\n\tTotal mem used:\t%d bytes\n\tNJAMD Overhead:\t%d bytes\n", user, tot_used, tot_used - user);
	
}
// vim:ts=4
