/* mem_over.c - Provides the overflow capabilities of NJAMD
 *
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 */
#include <lib/njamd.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <errno.h>

/* FUNCTION: nj_free_init 
 * SUMMARY: Performs the first half of free/realloc, namely retrieving an info
 *          pointer from the segment.
 */
static struct heap_entry *nj_free_init(void *start, void *userbuf, 
		u_long *laddr, char *caller)
{
	struct heap_entry *inf = NULL;
	
	TR;
	
	/* Check to see if we had to tack on another page in postmalloc */
	if(userbuf - start < NJ_ACCOUNTING_OVERHEAD)
	{
		start -= NJ_PAGE_SIZE;
	}

	/* FIXME: this may cause problems when known free differes from default,
	 * and system setup memory is actually freed. So far this hasn't presented
	 * itself */
	if(__nj_known_chk_free == ERROR_CHK_FREE)
	{
		if(mprotect(start, NJ_PAGE_SIZE, PROT_READ) 
				== -1)
		{
			TR;
			__nj_perror(__FUNCTION__"/mprotect");
			raise(SIGSEGV);
			return NULL;
		}
	
		TR;

		/* A NULL inf can only mean read from freed memory */
		if(*(struct heap_entry **)start == 0)
		{
			/* However, if the rest of the page isn't 0, we might have an
			 * underflow of 0 */
			if(memcmp(start, __nj_zero_page, NJ_PAGE_SIZE) != 0)
			{
				__nj_eprintf("NJAMD: heap corruption. Try using the underflow "
						"option to pinpoint source of error\n");	
				raise(SIGSEGV);
				return NULL;
			}

			__nj_eprintf("NJAMD/%s: Double %s of address 0x%lx\n", caller, caller, (u_long)userbuf);
			raise(SIGSEGV);
			return NULL;
		}
	}

	if(userbuf - start - NJ_ACCOUNTING_OVERHEAD >= sizeof(u_long) 
			&& *(u_long *)(start+NJ_ACCOUNTING_OVERHEAD) != NJ_FENCEPOST)
	{
		__nj_eprintf("NJAMD: heap corruption. Try using the underflow "
						"option to pinpoint source of error\n");	
		raise(SIGSEGV);
		return NULL;
	}
	TR;

	/* Index into the heap table */
	inf = *(struct heap_entry **)start;

	TR;

	if(!(__nj_heap_tbl <= inf && inf <= __nj_heap_top))
	{
		__nj_eprintf("NJAMD/%s: Heap corruption. Try using underflow option\n", caller);
		raise(SIGSEGV);
		exit(1);
	}


	if(inf->u_addr.start != start)
	{
		__nj_eprintf("NJAMD/%s: Inconsistancy. Either address passed (0x%lx) "
				"is not directly from malloc, or small underflow. Try using "
				"strict underflow option. \n", caller, userbuf);
		raise(SIGSEGV);
		return NULL;
	}


	TR;
	return inf;
}

/* FUNCTION: nj_free_fini
 * SUMMARY: performs the second half of free/realloc. Remaps memory pointed to
 *          by start as specified by the NJAMD_CHK_FREE setting.
 */

static void nj_free_fini(void *start, struct heap_entry *inf)
{
	int len = ROUND_TO_BLOCK(inf->ulen);
	int chk_free = __nj_known_chk_free;

	/* Check if allocated before we knew the protection scheme. So use the 
	 * preinit value (SEGV_CHK_FREE). */
	if((__nj_slab0 <= (u_long)start && (u_long)start < __nj_known_slab0))
		chk_free = PREINIT_CHK_FREE;

	/* We pass in negative arguments to subtract that lengh. But we only
	 * subtract the address space if NO_CHK_FREE is set. */
	__nj_update_stats(-inf->ulen, -ROUND_TO_FAULT(inf->ulen), (chk_free == NO_CHK_FREE) ? -ROUND_TO_BLOCK(inf->ulen) : 0);

	TR;
	switch(chk_free)
	{
		/* To check for double frees, we map a page of zero mem to clue us in */
		case ERROR_CHK_FREE:
			if(mmap(start, NJ_PAGE_SIZE, __nj_prot, MAP_FIXED | MAP_SHARED,
						__nj_zfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			if(mmap(start+NJ_PAGE_SIZE, len-NJ_PAGE_SIZE, __nj_prot, 
						MAP_FIXED | MAP_PRIVATE | MAP_ANON, __nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;

		case SEGV_CHK_FREE:
			if(mmap(start, len, __nj_prot, MAP_FIXED | MAP_PRIVATE | MAP_ANON,
						__nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;

		case NO_CHK_FREE:
			/* If we reuse the address space, we have to reuse the heap,
			 * otherwise it grows without bound */ 
			__nj_free_stack_push(inf);
			
			if(munmap(start, len) == -1)
			{
				__nj_perror(__FUNCTION__"/munmap");
				exit(1);
			}
			TR;
			break;

		case NO_FREE:
			/* Do nothing */
			break;

		default:
			__nj_eprintf("Unknown free checking option %d\n", chk_free);
			if(mmap(start, len, __nj_prot, MAP_FIXED | MAP_PRIVATE | MAP_ANON,
						__nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;
	}


	TR;
}

/* FUNCTION: __nj_overflow_alloc
 * SUMMARY: General purpose allocator, cuts memory off of the slab, protects
 *          the page after it. In order for the users memory to end on this
 *          page, we must offsett it as well.
 */
void *__nj_overflow_alloc(void *ptr, size_t len, int fill, u_long *laddr, char *caller)
{
	struct heap_entry *inf;
	size_t fetch, offset;
	void *buf;

	if(!len && NJ_ISSET(__nj_flags, NJ_ALLOW_MALLOC0))
	{
		return NULL;
	}

	/* Ask a stupid question, get a stupid answer */
	if(len <= 0)
	{
		__nj_eprintf("NJAMD/%s: Request must be > 0 (%d)\n", caller, len);
		__nj_print_stack_trace(laddr, TRACE_DEPTH);
		errno = EINVAL;
		return NULL;
	}

	/* By making len aligned by the specified alignment, the resulting buffer
	 * will be aligned */
		
	len = ALIGN_TO(len, __nj_align);

	/* We're going to need 2 extra pages of memory - one for any offset
	 * carryover over a page boundry, and one to protect. */
	fetch = ROUND_TO_BLOCK(len);
	/* The offset into our region where the user's memory starts */
	offset = fetch - len - NJ_PAGE_SIZE;
	
	TR;
	/* Realloc? */
	if(ptr != NULL)
	{
		void *start = (void *)((long)ptr & NJ_PAGE_MASK);
	
		TR;	
		if(!(inf = nj_free_init(start, ptr, laddr, caller)))
		{
			/* This should not happen */
			return NULL;
		}

		/* the actual start of the region is where the inf says */
		start = inf->u_addr.start;
		TR;
		
		/* Get a chunk from the slab allocator */
		buf = __nj_slab_fetch(fetch);
		
		/* copy the buffer and appropriate accounting info */	
		memcpy(buf + offset, ptr, NJ_MIN(inf->ulen, len));

		TR;
		/* Remap the old memory to bupkes */
		nj_free_fini(start, inf);

		if(__nj_known_chk_free != NO_CHK_FREE)
			memcpy(inf->loc.freed, laddr, TRACE_DEPTH*sizeof(u_long));

		*(struct heap_entry **)buf = inf = __nj_new_heap_entry(buf, len, laddr);
		start = buf;
		TR;

	}
	else
	{
		TR;
		buf = __nj_slab_fetch(fetch);

		if(fill != -1)
		{
			/* calloc */
			memset(buf, fill, fetch - NJ_PAGE_SIZE); 
		}	
		TR;
		*(struct heap_entry **)buf = inf = __nj_new_heap_entry(buf, len, laddr);
	}

	/* report the new memory */
	__nj_update_stats(len, fetch-NJ_PAGE_SIZE, fetch);

	inf->ulen = len;
	memcpy(inf->loc.alloced, laddr, TRACE_DEPTH*sizeof(u_long));

	if(mprotect(buf+offset+len, NJ_PAGE_SIZE, __nj_prot) == -1)
	{
		__nj_perror(__FUNCTION__"/mprotect");
		exit(1);
	}

	TR;

	/* To check for underflows */
	if(offset - NJ_ACCOUNTING_OVERHEAD >= sizeof(u_long))
	{
		*(u_long *)(buf+NJ_ACCOUNTING_OVERHEAD) = NJ_FENCEPOST;
	}

	return buf+offset;
}

/* FUNCTION: __nj_overflow_free
 * SUMMARY: Will unmap memory, then remap that address to anonymous memory with no
 * permissions.
 */ 
void __nj_overflow_free(void *buf, u_long *ret)
{
	void *start = (void *)((long)buf & NJ_PAGE_MASK);
	struct heap_entry *inf;

	if(buf == NULL)
	{
		return;
	}

	TR;	
	if(!(inf = nj_free_init(start, buf, ret, "free")))
	{
		/* Shouldn't happen */
		return;
	}

	memcpy(inf->loc.freed, ret, TRACE_DEPTH*sizeof(u_long));
	
	/* the actual start of the region is where inf says */
	start = inf->u_addr.start;

	nj_free_fini(start, inf);
}
// vim:ts=4
