/* mem_sunder.c - Provides the strict underflow capabilities of NJAMD
 *
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 * 
 */
#include <lib/njamd.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <errno.h>


/* FUNCTION: nj_free_init 
 * SUMMARY: Performs the first half of free/realloc, namely retrieving an info
 *          pointer from the segment.
 */
static struct heap_entry *nj_free_init(void *start, u_long *laddr, char *caller)
{
	struct heap_entry *inf;
	size_t remainder;	
	TR;

	/* Simple check to hint us to the user trying to pass an address not
	 * returned by malloc */
	if((u_long)start & ~(NJ_PAGE_MASK))
	{
		__nj_eprintf("NJAMD/%s: Address passed (0x%lx) is not page-aligned\n", 
				caller, (u_long)start, __nj_align);
		raise(SIGSEGV);
		return NULL;
	}

	TR;

	/* Since this is strict underflow, we must unprotect the first page to
	 * access the segment length. */
	if(mprotect(start, NJ_PAGE_SIZE, PROT_READ) == -1)
	{   
		TR;                                                   
		__nj_perror(__FUNCTION__"/mprotect");
		raise(SIGSEGV);
		return NULL;
	}

	if(__nj_known_chk_free == ERROR_CHK_FREE)
	{
		TR;
		if(*(struct heap_entry **)start == 0)
		{
			__nj_eprintf("NJAMD/%s: Double %s of address 0x%lx\n", caller, caller, (u_long)start);
			raise(SIGSEGV);
			return NULL;

		}
		TR;
	}

	inf = *(struct heap_entry **)start;

	remainder = NJ_PAGE_SIZE - (inf->ulen & ~NJ_PAGE_MASK); 

	if(remainder >= sizeof(u_long) 
			&& *(u_long *)(start+inf->ulen+NJ_PAGE_SIZE) != NJ_FENCEPOST)
	{
		__nj_eprintf("NJAMD: heap corruption. Try using the overflow "
						"option to pinpoint source of error\n");	
		raise(SIGSEGV);
		return NULL;
	}

	if(inf->u_addr.start != start)
	{
		__nj_eprintf("NJAMD/%s: Inconsistancy. Address passed (0x%lx) "
				"is not directly from malloc (or some pretty fierce heap "
			"corruption took place\n", caller, start);
		raise(SIGSEGV);
		return NULL;
	}

	TR;
	return inf;
}

/* FUNCTION: nj_free_fini
 * SUMMARY: performs the second half of free/realloc. Remaps memory pointed to
 *          by start as specified by the NJAMD_CHK_FREE setting.
 */
static void nj_free_fini(void *start, struct heap_entry *inf)
{
	/* start is acutally on the protected page. */
	int len = ROUND_TO_BLOCK(inf->ulen);
	int chk_free = __nj_known_chk_free;
	
	/* Check if allocated before we knew the protection scheme. So use the 
	 * preinit value (SEGV_CHK_FREE). */
	if((__nj_slab0 <= (u_long)start && (u_long)start < __nj_known_slab0))
		chk_free = PREINIT_CHK_FREE;

	/* Report the old memory as freed */
	__nj_update_stats(-inf->ulen, -ROUND_TO_BLOCK(inf->ulen), (chk_free == NO_CHK_FREE) ? -ROUND_TO_BLOCK(inf->ulen) : 0);
			
	switch(chk_free)
	{
		case ERROR_CHK_FREE:
			
			if(mmap(start, NJ_PAGE_SIZE, __nj_prot, MAP_FIXED | MAP_SHARED,
						__nj_zfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			if((len != NJ_PAGE_SIZE) && mmap(start+NJ_PAGE_SIZE, len-NJ_PAGE_SIZE, __nj_prot, 
						MAP_FIXED | MAP_PRIVATE | MAP_ANON, __nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;

		/* Unreliable for strict underflow */
		case SEGV_CHK_FREE:
			if(mmap(start, len, __nj_prot, MAP_FIXED | MAP_PRIVATE | MAP_ANON,
						__nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;

		case NO_CHK_FREE:
			__nj_free_stack_push(inf);
			
			if(munmap(start, len) == -1)
			{
				__nj_perror(__FUNCTION__"/munmap");
			}
			break;

		case NO_FREE:
			/* Do nothing */
			break;

		default:
			__nj_eprintf("Unknown free checking option %d\n", chk_free);

			if(mmap(start, len, __nj_prot, MAP_FIXED | MAP_PRIVATE | MAP_ANON,
						__nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;
	}

	
	TR;
}


/* FUNCTION: __nj_underflow_alloc
 * SUMMARY: General purpose allocator, cuts memory off of the slab, protects
 *          the page before it, and returns it.
 */
void *__nj_sunderflow_alloc(void *ptr, size_t len, int fill, u_long *laddr, char *caller)
{
	struct heap_entry *inf;
	size_t fetch, remainder;
	void *buf;

	TR;
	if(!len && NJ_ISSET(__nj_flags, NJ_ALLOW_MALLOC0))
	{
		return NULL;
	}

	/* Ask a stupid question, get a stupid answer */
	if(len <= 0)
	{
		__nj_eprintf("NJAMD/%s: Request must be > 0 (%d)\n", caller, len);
		__nj_print_stack_trace(laddr, TRACE_DEPTH);
		errno = EINVAL;
		return NULL;
	}

	len = ALIGN_TO(len, ALIGNMENT);

	/* We're going to need 2 extra pages of memory - one for any offset
	 * carryover over a page boundry, and one to protect. */
	fetch = ROUND_TO_BLOCK(len);
	/* The offset into our region where the user's memory starts */
	remainder = fetch - len - NJ_PAGE_SIZE;

	TR;
	/* Realloc? */
	if(ptr != NULL)
	{
		void *start = ptr - NJ_PAGE_SIZE;
		TR;	
		if(!(inf = nj_free_init(start, laddr, caller)))
		{
			/* This should not happen */
			return NULL;
		}
		
		TR;
		buf = __nj_slab_fetch(fetch);
		
		/* Copy the memory and info to the proper places */	
		memcpy(buf+NJ_PAGE_SIZE, ptr, NJ_MIN(inf->ulen, len));

		/* Remap the old memory to bupkes */
		nj_free_fini(start, inf);

		if(__nj_known_chk_free != NO_CHK_FREE)
			memcpy(inf->loc.freed, laddr, TRACE_DEPTH*sizeof(u_long));

		*(struct heap_entry **)buf = inf = __nj_new_heap_entry(buf, len, laddr);

		start = buf;
		TR;

	}
	else
	{
		TR;
		buf = __nj_slab_fetch(fetch);

		if(fill != -1)
		{
			memset(buf+NJ_PAGE_SIZE, fill, fetch - NJ_PAGE_SIZE); 
		}

		TR;

		*(struct heap_entry **)buf = inf = __nj_new_heap_entry(buf, len, laddr);
	}

	/* Report the new memory */
	__nj_update_stats(len, fetch, fetch);

	inf->ulen = len;
	memcpy(inf->loc.alloced, laddr, TRACE_DEPTH*sizeof(u_long));
		 
	if(mprotect(buf, NJ_PAGE_SIZE, __nj_prot) == -1)
	{
		__nj_perror(__FUNCTION__"/mprotect");
		exit(1);
	}

	TR;

	/* To check for overflows */
	if(remainder >= sizeof(u_long))
	{
		*(u_long *)(buf+NJ_PAGE_SIZE+len) = NJ_FENCEPOST;
	}

	return buf+NJ_PAGE_SIZE;
}

/* FUNCTION: __nj_underflow_free
 * SUMMARY: Will unmap memory, then remap that address to anonymous memory with no
 * permissions.
 */ 
void __nj_sunderflow_free(void *buf, u_long *ret)
{
	void *start = buf - NJ_PAGE_SIZE;
	struct heap_entry *inf;

	if(buf == NULL)
	{
		return;
	}

	TR;	
	if(!(inf = nj_free_init(start, ret, "free")))
	{
		/* Shouldn't happen */
		TR;
		return;
	}
	
	memcpy(inf->loc.freed, ret, TRACE_DEPTH*sizeof(u_long));

	nj_free_fini(start, inf);
}

/* FUNCTION: __nj_sunderflow_stats
 * SUMMARY: This is just a quick little stat function so I can see how much
 * memory is mapped. 
 */
void __nj_sunderflow_stats(void)
{
	struct heap_entry *p;
	int user = 0;
	int as_used = 0;
	int faulted = 0;
	
	for(p = __nj_heap_tbl; p < __nj_heap_top; p++)
	{
		if(p->ulen != 0)
		{
			if(!entry_is_freed(p))
			{
				user += p->ulen;
				faulted += ROUND_TO_BLOCK(p->ulen);
			}
			/* If the free protection isn't NO_CHK_FREE, then this memory is
			 * still mapped */
			if(__nj_known_chk_free != NO_CHK_FREE)
				as_used += ROUND_TO_BLOCK(p->ulen);
		}
	}

	__nj_eprintf("\nNJAMD totals:\n\tUser memory:\t\t%d bytes\n\tFaulted Memory:\t\t%d pages\n\tAddress space used:\t%d bytes\n\tNJAMD Overhead:\t\t%d bytes\n", user, faulted>>NJ_PAGE_SHIFT, as_used, faulted-user);
	
}
// vim:ts=4
