/* mem_under.c - Provides the weak underflow capabilities of NJAMD
 *
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 * 
 */
#include <lib/njamd.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <errno.h>



/* FUNCTION: nj_free_init 
 * SUMMARY: Performs the first half of free/realloc, namely retrieving an info
 *          pointer from the segment.
 */
static struct heap_entry *nj_free_init(void *start, u_long *laddr, char *caller)
{
	struct heap_entry *inf;
	size_t remainder;	
	TR;

	/* Simple check to hint us to the user trying to pass an address not
	 * returned by malloc */
	if((u_long)start & ~(NJ_PAGE_MASK))
	{
		__nj_eprintf("NJAMD/%s: Address passed (0x%lx) is not page-aligned\n", 
				caller, (u_long)start+NJ_OFFSET, __nj_align);
		raise(SIGSEGV);
		return NULL;
	}

	TR;

	if(__nj_known_chk_free == ERROR_CHK_FREE)
	{
		if(mprotect(start, NJ_PAGE_SIZE, PROT_READ) == -1)
		{
			TR;
			__nj_perror(__FUNCTION__"/mprotect");
			raise(SIGSEGV);
			return NULL;
		}

		TR;
		if(*(struct heap_entry **)start == 0)
		{
			/* if the size was zero, but the whole page isn't, we have an
 			 * underflow of 0.*/
			if(memcmp(start, __nj_zero_page, NJ_PAGE_SIZE) != 0)
			{
				TR;
				__nj_eprintf("NJAMD: heap corruption. Inconsistant allocation "
						"sizes. Try using the strict underflow option to " 
						" pinpoint source of error\n");	
				raise(SIGSEGV);
				return NULL;
			}
			__nj_eprintf("NJAMD/%s: Double %s of address 0x%lx\n"
					"0x%lx\n", caller, caller, (u_long)start + NJ_OFFSET);
			raise(SIGSEGV);
			return NULL;

		}
		TR;
	}

	inf = *(struct heap_entry **)start;

	if(!(__nj_heap_tbl <= inf && inf <= __nj_heap_top))
	{
		__nj_eprintf("NJAMD free: Heap corruption. Try using strict underflow option\n");
		raise(SIGSEGV);
		exit(1);
	}

	remainder = NJ_PAGE_SIZE - (ALIGN_TO(inf->ulen + NJ_OFFSET, ALIGNMENT) & ~NJ_PAGE_MASK); 
  
	if(remainder >= sizeof(u_long) 
			&& *(u_long *)(start+ALIGN_TO(inf->ulen + NJ_OFFSET, ALIGNMENT)) != NJ_FENCEPOST)
	{
		__nj_eprintf("NJAMD: heap corruption. Try using the overflow "
						"option to pinpoint source of error\n");	
		raise(SIGSEGV);
		return NULL;
	}

	if(inf->u_addr.start != start - NJ_PAGE_SIZE)
	{
		__nj_eprintf("NJAMD/%s: Inconsistancy. Either address passed (0x%lx) "
				"is not directly from malloc, or small underflow. Try using "
				"strict underflow option. \n", caller, start + NJ_OFFSET);
		raise(SIGSEGV);
		return NULL;
	}


	TR;
	return inf;
}

/* FUNCTION: nj_free_fini
 * SUMMARY: performs the second half of free/realloc. Remaps memory pointed to
 *          by start as specified by the NJAMD_CHK_FREE setting.
 */
static void nj_free_fini(void *start, struct heap_entry *inf)
{
	/* start is acutally AFTER the protected page. */
	int len = ROUND_TO_BLOCK(ALIGN_TO(inf->ulen + NJ_OFFSET, ALIGNMENT)) - NJ_PAGE_SIZE;
	int chk_free = __nj_known_chk_free;
	
	/* Check if allocated before we knew the protection scheme. So use the 
	 * preinit value (SEGV_CHK_FREE). */
	if((__nj_slab0 <= (u_long)start && (u_long)start < __nj_known_slab0))
		chk_free = PREINIT_CHK_FREE;

	/* We pass in negative arguments to subtract that lengh. But we only
	 * subtract the address space if NO_CHK_FREE is set. */
	__nj_update_stats(-inf->ulen, -(ROUND_TO_BLOCK(ALIGN_TO(inf->ulen + NJ_OFFSET, ALIGNMENT)) - NJ_PAGE_SIZE), (chk_free == NO_CHK_FREE) ? -ROUND_TO_BLOCK(ALIGN_TO(inf->ulen + NJ_OFFSET, ALIGNMENT)) : 0); 
		
	switch(chk_free)
	{
		case ERROR_CHK_FREE:
			
			if(mmap(start, NJ_PAGE_SIZE, __nj_prot, MAP_FIXED | MAP_SHARED,
						__nj_zfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			if((len != NJ_PAGE_SIZE) && mmap(start+NJ_PAGE_SIZE, len-NJ_PAGE_SIZE, __nj_prot, 
						MAP_FIXED | MAP_PRIVATE | MAP_ANON, __nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;

		case SEGV_CHK_FREE:
			if(mmap(start, len, __nj_prot, MAP_FIXED | MAP_PRIVATE | MAP_ANON,
						__nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;

		case NO_CHK_FREE:
			TR;
			__nj_free_stack_push(inf);
			
			if(munmap(start-NJ_PAGE_SIZE, len+NJ_PAGE_SIZE) == -1)
			{
				__nj_perror(__FUNCTION__"/munmap");
			}
			break;

		case NO_FREE:
			/* Do nothing */
			break;

		default:
			__nj_eprintf("Unknown free checking option %d\n", chk_free);

			if(mmap(start, len, __nj_prot, MAP_FIXED | MAP_PRIVATE | MAP_ANON,
						__nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;
	}

	TR;
}


/* FUNCTION: __nj_underflow_alloc
 * SUMMARY: General purpose allocator, cuts memory off of the slab, protects
 *          the page before it, and returns it.
 */
void *__nj_underflow_alloc(void *ptr, size_t len, int fill, u_long *laddr, char *caller)
{
	struct heap_entry *inf;
	size_t fetch, remainder;
	void *buf;
	size_t old_len = len;

	TR;
	if(!len && NJ_ISSET(__nj_flags, NJ_ALLOW_MALLOC0))
	{
		return NULL;
	}

	/* Ask a stupid question, get a stupid answer */
	if(len <= 0)
	{
		__nj_eprintf("NJAMD/%s: Request must be > 0 (%d)\n", caller, len);
		__nj_print_stack_trace(laddr, TRACE_DEPTH);
		errno = EINVAL;
		return NULL;
	}

	len = ALIGN_TO(len + NJ_OFFSET, ALIGNMENT);

	/* We're going to need 2 extra pages of memory - one for any offset
	 * carryover over a page boundry, and one to protect. */
	fetch = ROUND_TO_BLOCK(len);
	/* The offset into our region where the user's memory starts */
	remainder = fetch - len - NJ_PAGE_SIZE;

	TR;
	/* Realloc? */
	if(ptr != NULL)
	{
		void *start = ptr - NJ_OFFSET;
		TR;	
		if(!(inf = nj_free_init(start, laddr, caller)))
		{
			/* This should not happen */
			return NULL;
		}

		TR;
		
		buf = __nj_slab_fetch(fetch);
		
		/* Copy the memory and info to the proper places */	
		memcpy(buf+NJ_OFFSET+NJ_PAGE_SIZE, ptr, ALIGN_TO(NJ_MIN(inf->ulen, old_len) + NJ_OFFSET, ALIGNMENT)-NJ_OFFSET);

		/* Remap the old memory to bupkes */
		nj_free_fini(start, inf);

		if(__nj_known_chk_free != NO_CHK_FREE)
			memcpy(inf->loc.freed, laddr, TRACE_DEPTH*sizeof(u_long));

		*(struct heap_entry **)(buf + NJ_PAGE_SIZE) = inf = __nj_new_heap_entry(buf, old_len, laddr);

		start = buf;
		TR;
		

	}
	else
	{
		TR;
		buf = __nj_slab_fetch(fetch);

		/* Dont fault that first page */
		if(fill != -1)
		{
			memset(buf+NJ_PAGE_SIZE, fill, fetch - NJ_PAGE_SIZE); 
		}	
		TR;

		*(struct heap_entry **)(buf+NJ_PAGE_SIZE) = inf = __nj_new_heap_entry(buf, old_len, laddr);
	}

	__nj_update_stats(old_len, fetch-NJ_PAGE_SIZE, fetch);
	
	/* FIXME: What was so important about saving inf->ulen as old_len? 
	 * Why didn't I do this everywhere? */
	inf->ulen = old_len;
	memcpy(inf->loc.alloced, laddr, TRACE_DEPTH*sizeof(u_long));
		 
	if(mprotect(buf, NJ_PAGE_SIZE, __nj_prot) == -1)
	{
		__nj_perror(__FUNCTION__"/mprotect");
		exit(1);
	}

	TR;

	/* To check for overflows */	
	if(remainder >= sizeof(u_long))
	{
		*(u_long *)(buf+NJ_PAGE_SIZE+len) = NJ_FENCEPOST;
	}

	return buf+NJ_PAGE_SIZE+NJ_OFFSET;
}

/* FUNCTION: __nj_underflow_free
 * SUMMARY: Will unmap memory, then remap that address to anonymous memory with no
 * permissions.
 */ 
void __nj_underflow_free(void *buf, u_long *ret)
{
	void *start = buf - NJ_OFFSET;
	struct heap_entry *inf;

	if(buf == NULL)
	{
		return;
	}

	TR;	
	if(!(inf = nj_free_init(start, ret, "free")))
	{
		/* Shouldn't happen */
		TR;
		return;
	}

	memcpy(inf->loc.freed, ret, TRACE_DEPTH*sizeof(u_long));

	nj_free_fini(start, inf);
}

// vim:ts=4
