/* public.c - Provides the libc interface into NJAMD
 *
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 * 
 */

/* Avoid conflicting types */
#define pthread_create __nj_dummy1 
#define pthread_exit __nj_dummy2 
#define signal __nj_dummy4
#define exit __nj_dummy5

#include <lib/njamd.h>
#include <lib/libc.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>

#undef exit
#undef signal
#undef pthread_exit
#undef pthread_create

#ifdef HAVE_SANE_DLOPEN
/* FUNCTION: signal
 * SUMMARY: Wraps libc signal, makes sure the user doesn't overwrite sigusr1.
 *          If they do, store their handler to execute after ours. 
 */
nj_sighandler_t signal(int num, nj_sighandler_t newhandler)
{
	nj_sighandler_t ret;
	TR;

	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}
	TR;

# ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
# endif

	if(!NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY) 
			&& NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		__nj_secondary_init();
	}
	TR;

	/* If we dump hardcore, we don't handle segv */
	if(NJ_ISSET(__nj_flags, NJ_DUMP_HARDCORE))
	{
		switch(num)
		{
			case SIGQUIT:
			case SIGABRT:
			case SIGUSR1:
			case SIGINT:
			case SIGTERM:
				ret = __nj_sig_hdlr[num];
				__nj_sig_hdlr[num] = newhandler;
				__nj_sig_actn[num] = NULL;
				return ret;
		}

	}
	else
	{
		switch(num)
		{
			case SIGQUIT:
			case SIGILL:
			case SIGABRT:
			case SIGFPE:
			case SIGUSR1:
			case SIGSEGV:
			case SIGBUS:
			case SIGINT:
			case SIGTERM:
				ret = __nj_sig_hdlr[num];
				__nj_sig_hdlr[num] = newhandler;
				__nj_sig_actn[num] = NULL;
				return ret;
		}
	}
	TR;
	if(__njLibcSignal)
		return __njLibcSignal(num, newhandler);
	else
	{
		__nj_eprintf("Error!!! Undefined symbol signal\n");
		errno = ENOSYS;
		return NULL;
	}
}

# ifdef HAVE_POSIX_SIGACTION
/* FUNCTION: sigaction
 * SUMMARY: Wraps libc sigaction, makes sure the user doesn't overwrite sigusr1.
 *          If they do, store their handler to execute after ours. 
 */
int sigaction(int num, const struct sigaction *new, struct sigaction *old)
{
	
	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}
	TR;

#  ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
#  endif

	if(!NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY) 
			&& NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		__nj_secondary_init();
	}
	TR;
	
	/* Handle deadly signals internally, so we always exit cleanly and with
	 * status messages, if the user wishes */
	if(NJ_ISSET(__nj_flags, NJ_DUMP_HARDCORE))
	{
		switch(num)
		{
			case SIGUSR1:
			case SIGQUIT:
			case SIGABRT:
			case SIGINT:
			case SIGTERM:
				TR;
				if(old)
				{
					old->sa_handler = __nj_sig_hdlr[num];
					old->sa_sigaction = __nj_sig_actn[num];
				}
				if(new)
				{
					if(new->sa_flags & SA_SIGINFO)
					{
						__nj_sig_actn[num] = new->sa_sigaction;
						__nj_sig_hdlr[num] = SIG_DFL;
					}
					else
					{
						__nj_sig_hdlr[num] = new->sa_handler;
						__nj_sig_actn[num] = NULL;
					}
				}
				return 0;
		}
	}
	else
	{
		switch(num)
		{
			case SIGUSR1:
			case SIGILL:
			case SIGFPE:
			case SIGQUIT:
			case SIGABRT:
			case SIGSEGV:
			case SIGBUS:
			case SIGINT:
			case SIGTERM:
				TR;
				if(old)
				{
					old->sa_handler = __nj_sig_hdlr[num];
					old->sa_sigaction = __nj_sig_actn[num];
				}
				if(new)
				{
					if(new->sa_flags & SA_SIGINFO)
					{
						__nj_sig_actn[num] = new->sa_sigaction;
						__nj_sig_hdlr[num] = SIG_DFL;
					}
					else
					{
						__nj_sig_hdlr[num] = new->sa_handler;
						__nj_sig_actn[num] = NULL;
					}
				}
				return 0;
		}
	}
	TR;
	if(__njLibcSigaction)
		return __njLibcSigaction(num, new, old);
	else
	{
		__nj_eprintf("Error!!! Undefined symbol sigaction\n");
		errno = ENOSYS;
		return -1;
	}
}
# endif /* SIGACTION */

# ifdef _THREAD_SAFE
/* FUNCTION: pthread_create
   SUMMARY: We wrap pthread_create so we KNOW that __nj_primary_init will run
   before any threads exist.
 */
int  pthread_create(pthread_t  *  thread, pthread_attr_t *attr,
		void * (*start_routine)(void *), void * arg)
{
	static struct pthread_arg_wrapper warg;
	int ret;

	TR;
	NJ_SET(__nj_flags, NJ_ACTIVE_PTHREADS);

	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}
	TR;

#  ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
#  endif

	if(!NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY) && NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		__nj_secondary_init();
	}
	TR;
	
	pthread_mutex_lock(&__nj_pthread_lock);

	warg.start = start_routine;
	warg.real_arg = arg;
	
	if(__nj_pthread_create)
		ret = __nj_pthread_create(thread, attr, __nj_pthread_launch, &warg);
	else
	{
		__nj_eprintf("Error!!! Undefined symbol pthread_create\n");
		ret = -1;
	}

   /* This unlock moved to __nj_pthread_launch to fix a race condition spotted
    * by Aleks Totic
	pthread_mutex_unlock(&__nj_pthread_lock); */

	return ret;
}

/* Food for the dynamic linker. Do NOT move.  */
void __nj_pthread_exit_start()
{
	return;
}

void pthread_exit(void *retval)
{

	/* Pthread exit damn well better not be called before create */
	if(__nj_pthread_exit)
	{
		__nj_pthread_exit(retval);
	}
	else
	{
		__nj_eprintf("Error!!! Undefined symbol pthread_exit\n");
	}

}
# endif /* THREADSAFE */
#endif /* DLOPEN */

/* Do NOT move this from pthread_exit */
void __nj_pthread_exit_end(void)
{
	return;
}

#ifdef HAVE_SANE_DLOPEN

/* Food for the dynamic linker. Do NOT move.  */
void __nj_exit_start(void)
{
	return;
}

void exit(int ret)
{
	if(!NJ_ISSET(__nj_flags, NJ_RAN_LIBC))
	{
		__njLibcFuncInit();
	}
	
	if(__nj_exit)
	{
		__nj_exit(ret);
	}

	__nj_eprintf("Error!!! Undefined symbol exit\n");
	nj_exit(ret);
	
}
#endif

/* Do NOT move this from exit() */
void __nj_exit_end(void)
{
	return;
}
/* FUNCTION: malloc 
 * SUMMARY: Wraps libc malloc, running init code if needed.  */
void *malloc(size_t size)
{
	u_long ret_addrs[TRACE_DEPTH];

	TR;
	/* FIXME: pipeline flush */
	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}

	TR;

#ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
#endif

	__nj_userspace_ret(ret_addrs, TRACE_DEPTH);

	/* If the allocator type is unknown (ie the seconday 
	   init hasn't been run yet), use the default one */
	if(NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY))
	{
		TR;
		return __nj_known_alloc(NULL, size, -1, ret_addrs, __FUNCTION__);
	}
	else if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		TR;
		return PREINIT_ALLOCATOR(NULL, size, -1, ret_addrs, __FUNCTION__);
	}
	else /* we are in main, but didn't run init yet */
	{
		TR;
		__nj_secondary_init();
		return __nj_known_alloc(NULL, size, -1, ret_addrs, __FUNCTION__);
	}
}

/* FUNCTION: valloc 
 * SUMMARY: Wraps libc valloc, which simply calls malloc.  */
void *valloc(size_t size)
{
	return malloc(size);
}

/* FUNCTION: calloc 
 * SUMMARY: Wraps libc calloc, calling init code when needed */
void *calloc(size_t nmemb, size_t size)
{
	u_long ret_addrs[TRACE_DEPTH];

	TR;
	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}

	TR;
#ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
#endif
	
	__nj_userspace_ret(ret_addrs, TRACE_DEPTH);

	if(NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY))
	{
		return __nj_known_alloc(NULL, size*nmemb, 0, ret_addrs, __FUNCTION__);
	}
	else if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		return PREINIT_ALLOCATOR(NULL, size*nmemb, 0, ret_addrs, __FUNCTION__);
	}
	else /* we are in main, but didn't run init yet */
	{
		__nj_secondary_init();
		return __nj_known_alloc(NULL, size*nmemb, 0, ret_addrs, __FUNCTION__);
	}
	
}

/* FUNCTION: realloc 
 * SUMMARY: Wraps libc realloc, calling free if size is 0 */
void *realloc(void *ptr, size_t size)
{
	u_long ret_addrs[TRACE_DEPTH];
	
	TR;	
	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}

	TR;

#ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
#endif

	__nj_userspace_ret(ret_addrs, TRACE_DEPTH);

	if(NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY))
	{
known:
		if(size != 0)
		{
			return __nj_known_alloc(ptr, size, -1, ret_addrs, __FUNCTION__);
		}
		else
		{
			/* if ptr is from a section of memory allocated before we knew which
			 * allocator to use, free it with default */
			if(!((u_long)ptr >= __nj_slab0 && (u_long)ptr < __nj_known_slab0))
				__nj_known_free(ptr, ret_addrs);
			else
				PREINIT_DEALLOCATOR(ptr, ret_addrs);
			return NULL;
		}
	}
	else if(NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		__nj_secondary_init();
		goto known;
	}

	if(size != 0)
	{
		/* must write realloc pair using mremap */
		return PREINIT_ALLOCATOR(ptr, size, -1, ret_addrs, __FUNCTION__);
	}
	else 
	{
		PREINIT_DEALLOCATOR(ptr, ret_addrs);
		return NULL;
	}

}

/* FUNCTION: free 
 * SUMMARY: Wraps libc free. We keep track of the region of memory allocated
 * with an unknown allocator. */
void free(void *ptr)
{
	u_long ret_addrs[TRACE_DEPTH];

	if(!ptr && NJ_ISSET(__nj_flags, NJ_ALLOW_FREE0))
	{
		return;
	}
	
	__nj_userspace_ret(ret_addrs, TRACE_DEPTH);

	TR;
	/* We can only perform this check if protection is NOT no_chk_free. Cause
	 * in that case, we start to go outside the slabs */
	if(ptr && __nj_address_from_malloc(ptr))
	{
		TR;
		/* if ptr is from a section of memory allocated before we knew which
		 * allocator to use, free it with default */
		if(NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY) 
				&& !(__nj_slab0 <= (u_long)ptr && (u_long)ptr < __nj_known_slab0))
		{
			TR;
			__nj_known_free(ptr, ret_addrs);
		}
		else /* We don't run init from free. It's not nessicary */
			PREINIT_DEALLOCATOR(ptr, ret_addrs);
	}
	else
	{
		__nj_eprintf("NJAMD/free: Address 0x%x not from malloc\n", ptr);
		__nj_print_stack_trace(ret_addrs, TRACE_DEPTH);
		return;
	}
}
// vim:ts=4
