#include <lib/njamd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

/* Portability */
#ifndef NJ_PAGE_SIZE
# if defined(DEBUG) || defined(EDBUG)
#  warning "**** Page Size undefined ****"
# endif
u_long 					NJ_PAGE_SIZE;
#endif

/* */
#ifndef NJ_PAGE_MASK
u_long 					NJ_PAGE_MASK;
#endif

#ifndef NJ_PAGE_SHIFT
u_long 					NJ_PAGE_SHIFT;
#endif

struct heap_entry zero_entry; /* for refrence */
struct mem_stats mem_usage;

u_int __nj_leaked_user;
u_int __nj_leaked_faulted;

void show_usage(char *prog)
{
		fprintf(stderr, "Usage: %s [options] <heap file>\n", prog);
		fprintf(stderr, " -a <address>      Display information known about <address>.\n");
		fprintf(stderr, " -d <depth>        Display stack trace (up to %d)\n",
				TRACE_DEPTH);
		fprintf(stderr, " -t                Trim heap file to actual size\n");
		fprintf(stderr, " -l                Dump leaked memory.\n");
		fprintf(stderr, " -f                Dump freed memory.\n");
		fprintf(stderr, " -h                Show this help.\n");
}

int trim_heap(struct heap_entry *tbl, int fd, int entries)
{
	int i;
	for(i=0; i < entries; i++)
	{
		if(memcmp(&tbl[i], &zero_entry, sizeof(zero_entry)) == 0)
		{
			ftruncate(fd, i*sizeof(struct heap_entry));
			return i;
		}
	}
	return entries;
}

struct heap_entry *init_heap(char *file, size_t *size, int trim)
{
	int fd;
	struct stat st;
	struct heap_entry *map;
	int new_size;

	if((fd = open(file, O_RDWR)) == -1)
	{
		perror("open");
		exit(1);
	}

	fstat(fd, &st);

	if((map = mmap(0, st.st_size, PROT_READ, MAP_SHARED, fd, 0)) == (void *)-1)
	{
		perror("mmap");
		exit(1);
	}

	if(size)
		*size = (st.st_size - sizeof(struct mem_stats)) / sizeof(struct heap_entry);

	if(trim)
	{
		new_size = trim_heap(map, fd, *size);

		if(new_size != *size)
		{
			/* If we trimmed the heap and the size changed, we probably have too
			 * many entries */
			*size = (new_size*sizeof(struct heap_entry) - sizeof(struct mem_stats)) / sizeof(struct heap_entry);
		}
	}
	mem_usage = *((struct mem_stats *)&map[*size]); 
	
	close(fd);
	
	return map;	
}

struct heap_entry *find_addr(struct heap_entry *tbl, size_t size, char *addr)
{
	u_long address = 0;
	int i;
	
	errno = 0;
	address = strtoul(addr, NULL, 0);
	
	if(errno)
	{
		perror("strtoul");
		exit(1);
	}

	for(i=0; i < size; i++)
	{
		/* Make sure ulen is not zero */
		if(tbl[i].ulen && (u_long)tbl[i].u_addr.start < address && address <
				(u_long)tbl[i].u_addr.start + ROUND_TO_BLOCK(tbl[i].ulen))
		{
			return &tbl[i];
		}
	}

	return NULL;
	
}

void __nj_print_stack_trace(u_long *addr_array, int depth)
{
	int i;

	/* Change in NJAMDPM */
	if(addr_array[0] == NJ_RETADDR_NONE)
	{
		printf("\tCall stack not saved\n");
		return;
	}

	for(i=0; i < depth; i++)
	{
		if(addr_array[i] == 0)
			return;
		else if(addr_array[i] != NJ_DESTRUCTOR)
			printf("\tcalled from 0x%lx\n", addr_array[i]);
		else
			printf("\tcalled from a destructor during program exit\n");
	}
	
}

int dump_entry(struct heap_entry *entry, int depth)
{
	if(!entry->ulen)
	{
		printf("\n0x%lx-0x%lx: Freed Memory\n", (u_long)entry->u_addr.start,
				(u_long)entry->u_addr.start + ROUND_TO_BLOCK(entry->ulen));
	}

	/* Change in NJAMDPM */
	if(!entry->loc.alloced[0])
	{
		TR;
		return 0;
	}

	printf("\n0x%lx-0x%lx: Aligned len %d\n   Allocation callstack:\n", 
			(u_long)entry->u_addr.start, 
			(u_long)entry->u_addr.start + ROUND_TO_BLOCK(entry->ulen), 
			entry->ulen);

	__nj_print_stack_trace(entry->loc.alloced, depth);
	
	if(!entry_is_freed(entry))
		printf("   Not Freed\n");
	else
	{
		printf("   Freed callstack:\n");
		__nj_print_stack_trace(entry->loc.freed, depth);
	}

	return 1;
}

void dump_table(struct heap_entry *tbl, int size, int freed, int leaked, int
		depth)
{
	int i;
	unsigned int free_cnt = 0;
	
	if(!(leaked || freed))
		return;
		
	printf("\nHeap statistics:\n");
	
	for(i=0; i < size; i++)
	{
		if(tbl[i].ulen != 0)
		{
			if(freed && entry_is_freed(&tbl[i]))
			{
				if(dump_entry(&tbl[i], depth))
					free_cnt += tbl[i].ulen;
			}
			else if(leaked && !entry_is_freed(&tbl[i]))
			{
				if(dump_entry(&tbl[i], depth))
				{
					__nj_leaked_user += tbl[i].ulen;
					switch(mem_usage.alloc_type)
					{
						case ALLOC_OVER:
							__nj_leaked_faulted += ROUND_TO_PAGE(tbl[i].ulen);
							break;
						case ALLOC_UNDER:
							__nj_leaked_faulted += ROUND_TO_PAGE(ALIGN_TO(tbl[i].ulen + NJ_OFFSET, ALIGNMENT));
							break;
						case ALLOC_SUNDER:
							__nj_leaked_faulted += ROUND_TO_BLOCK(tbl[i].ulen);
						case ALLOC_NONE:
							break;
					}
				}
			}
		}
	}

	puts("");

	if(freed)
		printf("Aligned freed memory: %u\n", free_cnt);
	if(leaked)
		printf("Aligned leaked memory: %u\n", __nj_leaked_user);
	
}

char *__nj_byte_to_string(char *buf, int len, int bytes)
{
	if(bytes > 1<<SHIFT_MEGABYTE)
		snprintf(buf, len, "%.3f mB", bytes/(1024*1024.0));
	else if(bytes > 1024)
		snprintf(buf, len, "%.3f kB", bytes/1024.0);
	else
		snprintf(buf, len, "%d bytes", bytes);

	return buf;
}

/* FUNCTION: __nj_usage_stats
 * SUMMARY: This is just a quick little stat function so I can see how much
 * memory is mapped. 
 */
void dump_stats(void)
{
	/* Kinda silly, but what the hell. eprintf is much more expensive than
	 * stack space allocation. */
	char buf1[80], buf2[80], buf3[80], buf4[80], buf5[80], buf6[80], buf7[80];

	if(!mem_usage.total_allocs)
	{
		printf("\nNJAMD Totals:\n\tmalloc() was never called!\n");
		return;
	}

	printf(
			"\nNJAMD totals:\n\n"
			"\tAllocation totals:\t\t%d total, %d leaked\n"
			"\tLeaked User Memory:\t\t%s\n"
			"\tPeak User Memory:\t\t%s\n"
			"\tNJAMD Overhead at peak:\t\t%s\n" 
			"\tPeak NJAMD Overhead:\t\t%s\n" 
			"\tAverage NJAMD Overhead:\t\t%s per alloc\n" 
			"\tAddress space used:\t\t%s\n"
			"\tNJAMD Overhead at exit:\t\t%s\n\n",
			mem_usage.total_allocs, mem_usage.leaked_allocs, 
			__nj_byte_to_string(buf1, sizeof(buf1), __nj_leaked_user), 
			__nj_byte_to_string(buf2, sizeof(buf2), mem_usage.peak_user), 
			__nj_byte_to_string(buf3, sizeof(buf3), mem_usage.peak_faulted-mem_usage.peak_user),
			__nj_byte_to_string(buf4, sizeof(buf5), mem_usage.peak_overhead),
			__nj_byte_to_string(buf5, sizeof(buf5), (mem_usage.total_faulted-mem_usage.total_user)/mem_usage.total_allocs),
			__nj_byte_to_string(buf6, sizeof(buf6), mem_usage.as_used), 
			__nj_byte_to_string(buf7, sizeof(buf7), __nj_leaked_faulted-__nj_leaked_user)
				);
}

/* What we need to do:
 * 1. Take an address arguement and find it
   2. Dump leaked memory
   3. Dump freed memory 
   */
int main(int argc, char **argv)
{
	struct heap_entry *heap_tbl, *inf;
	char *heap_file = NULL;
	int display_stats = 0;
	int display_leaks = 0;
	int display_freed = 0;
	int trim_file = 0;
	char *addrstr = NULL;
	struct stat st;
	size_t entries;
	int i;
	int depth=TRACE_DEPTH;
	
	if(argc < 3)
	{
		show_usage(argv[0]);
		exit(1);
	}
	
	while((i = getopt(argc, argv, "a:d:thsfl")) != -1)
	{		
		switch(i)
		{
			case 'a':
				addrstr = optarg;
				break;
			case 'd':
				errno=0;
				depth = strtol(optarg, NULL, 0);
				if(errno || depth > TRACE_DEPTH)
				{
					fprintf(stderr, "invalid depth %d\n", depth);
					exit(1);
				}
				break;
			case 't':
				trim_file=1;
				break;
			case 'h':
				show_usage(argv[0]);
				exit(1);
				break;
			case 'f':
				display_freed = 1;
				break;
			case 's':
				display_stats = 1;
				break;
			case 'l':
				display_leaks = 1;
				break;
		}
	}
	
	if(optind >= argc)
	{
		show_usage(argv[0]);
		exit(1);
	}
			
	heap_file = argv[optind];

	if(stat(heap_file, &st) == -1)
	{
		perror("heap file");
		exit(1);
	}

#ifndef NJ_PAGE_SIZE
	NJ_PAGE_SIZE = getpagesize();
#endif
	
#ifndef NJ_PAGE_MASK
	NJ_PAGE_MASK = ~(NJ_PAGE_SIZE-1);
#endif
	
#ifndef NJ_PAGE_SHIFT
	for(i=0; i < sizeof(u_long)*8; i++)
	{
		if(NJ_PAGE_SIZE & (1<<i))
		{
			NJ_PAGE_SHIFT = i;
			break;
		}
	}
#endif
	
	heap_tbl = init_heap(heap_file, &entries, trim_file);
	
	
	if(addrstr)
	{
		printf("\nStats on address %s:\n", addrstr);
		inf = find_addr(heap_tbl, entries, addrstr);
		if(!inf)
			fprintf(stderr, "Address %s not in the heap table.\n", addrstr);
		else
			dump_entry(inf, depth);
	}

	dump_table(heap_tbl, entries, display_freed, display_leaks, depth);	

	if(display_stats)
		dump_stats();

	munmap(heap_tbl, entries*sizeof(struct heap_entry));

	exit(0);
}
// vim:ts=4
